@@ -39,22 +39,23 @@
 #include "chunkserver/hddspacemgr.h"
 #include "chunkserver/network_main_thread.h"
 #include "common/cfg.h"
-#include "protocol/cstoma.h"
 #include "common/datapack.h"
 #include "common/goal.h"
-#include "protocol/input_packet.h"
+#include "common/loop_watchdog.h"
 #include "common/main.h"
 #include "common/massert.h"
-#include "protocol/matocs.h"
-#include "protocol/MFSCommunication.h"
 #include "common/moosefs_vector.h"
 #include "common/output_packet.h"
-#include "protocol/packet.h"
 #include "common/random.h"
 #include "common/slogger.h"
 #include "common/sockets.h"
 #include "common/time_utils.h"
 #include "devtools/request_log.h"
+#include "protocol/cstoma.h"
+#include "protocol/input_packet.h"
+#include "protocol/matocs.h"
+#include "protocol/MFSCommunication.h"
+#include "protocol/packet.h"
 
 #define MaxPacketSize 10000
 
@@ -395,12 +396,19 @@ void masterconn_replicate(const std::vector<uint8_t>& data) {
 	matocs::replicateChunk::deserializePartial(data,
 			chunkId, chunkVersion, chunkType, sourcesBuffer);
 	sourcesBufferSize = data.size() - (sourcesBuffer - data.data());
+
 	OutputPacket* outputPacket = new OutputPacket;
 	cstoma::replicateChunk::serialize(outputPacket->packet,
 			chunkId, chunkType, LIZARDFS_STATUS_OK, chunkVersion);
 	DEBUG_LOG("cs.matocs.replicate") << chunkId;
-	job_replicate(jpool, masterconn_lizjobfinished, outputPacket,
-			chunkId, chunkVersion, chunkType, sourcesBufferSize, sourcesBuffer);
+	if (hdd_scans_in_progress()) {
+		// Folder scan in progress - replication is not possible
+		masterconn_lizjobfinished(LIZARDFS_ERROR_WAITING, outputPacket);
+	} else {
+		job_replicate(jpool, masterconn_lizjobfinished, outputPacket,
+				chunkId, chunkVersion, chunkType, sourcesBufferSize, sourcesBuffer);
+	}
+
 }
 
 void masterconn_legacy_replicate(masterconn *eptr,const uint8_t *data,uint32_t length) {
@@ -423,7 +431,10 @@ void masterconn_legacy_replicate(masterconn *eptr,const uint8_t *data,uint32_t l
 	ptr = masterconn_get_packet_data(packet);
 	put64bit(&ptr,chunkid);
 	put32bit(&ptr,version);
-	if (length==8+4+4+2) {
+
+	if (hdd_scans_in_progress()) {
+		masterconn_replicationfinished(LIZARDFS_ERROR_WAITING, packet);
+	} else if (length==8+4+4+2) {
 		ip = get32bit(&data);
 		port = get16bit(&data);
 //              syslog(LOG_NOTICE,"start job replication (%08" PRIX64 ":%04" PRIX32 ":%04" PRIX32 ":%02" PRIX16 ")",chunkid,version,ip,port);
@@ -661,6 +672,9 @@ void masterconn_connecttest(masterconn *eptr) {
 }
 
 void masterconn_read(masterconn *eptr) {
+	ActiveLoopWatchdog watchdog(std::chrono::milliseconds(20));
+
+	watchdog.start();
 	for (;;) {
 		if (job_pool_jobs_count(jpool) >= (BGJOBSCNT * 9) / 10) {
 			return;
@@ -696,11 +710,18 @@ void masterconn_read(masterconn *eptr) {
 
 		masterconn_gotpacket(eptr, eptr->inputPacket.getHeader(), eptr->inputPacket.getData());
 		eptr->inputPacket.reset();
+
+		if (watchdog.expired()) {
+			break;
+		}
 	}
 }
 
 void masterconn_write(masterconn *eptr) {
+	ActiveLoopWatchdog watchdog(std::chrono::milliseconds(20));
 	int32_t i;
+
+	watchdog.start();
 	while (!eptr->outputPackets.empty()) {
 		OutputPacket& pack = eptr->outputPackets.front();
 		i=write(eptr->sock, pack.packet.data() + pack.bytesSent,
@@ -718,6 +739,10 @@ void masterconn_write(masterconn *eptr) {
 			return;
 		}
 		eptr->outputPackets.pop_front();
+
+		if (watchdog.expired()) {
+			break;
+		}
 	}
 }
 
