@@ -33,19 +33,20 @@
 #include <deque>
 
 #include "common/chunks_availability_state.h"
+#include "common/chunk_copies_calculator.h"
 #include "common/datapack.h"
 #include "common/exceptions.h"
 #include "common/flat_set.h"
 #include "common/goal.h"
 #include "common/hashfn.h"
 #include "common/lizardfs_version.h"
+#include "common/loop_watchdog.h"
 #include "common/main.h"
 #include "common/massert.h"
 #include "common/slice_traits.h"
 #include "common/small_vector.h"
 #include "protocol/MFSCommunication.h"
 #include "master/checksum.h"
-#include "master/chunk_copies_calculator.h"
 #include "master/chunk_goal_counters.h"
 #include "master/filesystem.h"
 #include "master/goal_cache.h"
@@ -250,6 +251,24 @@ public:
 	static GoalCache goalCache;
 #endif
 
+	void clear() {
+		goalCounters_.clear();
+		next = nullptr;
+		chunkid = 0;
+		version = 0;
+		lockid = 0;
+		lockedto = 0;
+		checksum = 0;
+#ifndef METARESTORE
+		inEndangeredQueue = 0;
+		needverincrease = 1;
+		interrupted = 0;
+		operation = NONE;
+		slisthead = NULL;
+		initStats();
+#endif
+	}
+
 	// Highest id of the chunk's goal
 	// This function is preserved only for backward compatibility of metadata checksums
 	// and shouldn't be used anywhere else.
@@ -795,6 +814,7 @@ static inline chunk* chunk_malloc() {
 	if (gChunksMetadata->chfreehead) {
 		ret = gChunksMetadata->chfreehead;
 		gChunksMetadata->chfreehead = ret->next;
+		ret->clear();
 		return ret;
 	}
 	if (gChunksMetadata->cbhead==NULL || gChunksMetadata->cbhead->firstfree==CHUNK_BUCKET_SIZE) {
@@ -805,6 +825,7 @@ static inline chunk* chunk_malloc() {
 	}
 	ret = (gChunksMetadata->cbhead->bucket)+(gChunksMetadata->cbhead->firstfree);
 	gChunksMetadata->cbhead->firstfree++;
+	ret->clear();
 	return ret;
 }
 
@@ -823,19 +844,8 @@ chunk* chunk_new(uint64_t chunkid, uint32_t chunkversion) {
 	gChunksMetadata->chunkhash[chunkpos] = newchunk;
 	newchunk->chunkid = chunkid;
 	newchunk->version = chunkversion;
-	newchunk->lockid = 0;
-	newchunk->lockedto = 0;
-#ifndef METARESTORE
-	newchunk->inEndangeredQueue = 0;
-	newchunk->needverincrease = 1;
-	newchunk->interrupted = 0;
-	newchunk->operation = NONE;
-	newchunk->slisthead = NULL;
-	newchunk->initStats();
-#endif
 	gChunksMetadata->lastchunkid = chunkid;
 	gChunksMetadata->lastchunkptr = newchunk;
-	newchunk->checksum = 0;
 	chunk_update_checksum(newchunk);
 	return newchunk;
 }
@@ -1285,7 +1295,7 @@ uint8_t chunk_multi_truncate(uint64_t ochunkid, uint32_t lockid, uint32_t length
 	}
 	if (denyTruncatingParityParts) {
 		for (slist *s = oc->slisthead; s; s = s->next) {
-			if (slice_traits::xors::isXorParity(s->chunkType)) {
+			if (slice_traits::isParityPart(s->chunkType)) {
 				return LIZARDFS_ERROR_NOTPOSSIBLE;
 			}
 		}
@@ -1507,10 +1517,11 @@ const ChunksAvailabilityState& chunk_get_availability_state(bool regularChunksOn
 
 struct ChunkLocation {
 	ChunkLocation() : chunkType(slice_traits::standard::ChunkPartType()),
-			distance(0), random(0) {
+			chunkserver_version(0), distance(0), random(0) {
 	}
 	NetworkAddress address;
 	ChunkPartType chunkType;
+	uint32_t chunkserver_version;
 	uint32_t distance;
 	uint32_t random;
 	MediaLabel label;
@@ -1549,6 +1560,7 @@ int chunk_getversionandlocations(uint64_t chunkid, uint32_t currentIp, uint32_t&
 					&(chunkserverLocation.address.port),
 					&(chunkserverLocation.label)) == 0) {
 				chunkserverLocation.chunkType = s->chunkType;
+				chunkserverLocation.chunkserver_version = matocsserv_get_version(s->ptr);
 				chunkserverLocation.distance =
 						topology_distance(chunkserverLocation.address.ip, currentIp);
 						// in the future prepare more sophisticated distance function
@@ -1561,7 +1573,7 @@ int chunk_getversionandlocations(uint64_t chunkid, uint32_t currentIp, uint32_t&
 	std::sort(chunkLocation.begin(), chunkLocation.end());
 	for (uint32_t i = 0; i < chunkLocation.size(); ++i) {
 		const ChunkLocation& loc = chunkLocation[i];
-		serversList.emplace_back(loc.address, loc.chunkType);
+		serversList.emplace_back(loc.address, loc.chunkType, loc.chunkserver_version);
 	}
 	return LIZARDFS_STATUS_OK;
 }
@@ -1935,6 +1947,7 @@ public:
 private:
 	typedef std::vector<ServerWithUsage> ServersWithUsage;
 
+	uint32_t getMinChunkserverVersion(chunk *c, ChunkPartType type);
 	bool tryReplication(chunk *c, ChunkPartType type, matocsserventry *destinationServer);
 
 	void deleteInvalidChunkParts(chunk *c);
@@ -2023,48 +2036,70 @@ static bool chunkPresentOnServer(chunk *c, Goal::Slice::Type slice_type, matocss
 	return false;
 }
 
+uint32_t ChunkWorker::getMinChunkserverVersion(chunk */*c*/, ChunkPartType /*type*/) {
+	return kFirstECVersion;
+}
+
 bool ChunkWorker::tryReplication(chunk *c, ChunkPartType part_to_recover,
-				matocsserventry *destinationServer) {
+				matocsserventry *destination_server) {
 	// TODO(msulikowski) Prefer VALID over TDVALID copies.
-	// NOTE: we don't allow replicating xor chunks from pre-xor chunkservers
 	std::vector<matocsserventry *> standard_servers;
-	std::vector<matocsserventry *> xor_capable_servers;
-	std::vector<ChunkPartType> xor_capable_parts;
-	ChunkCopiesCalculator xor_capable_calc(c->getGoal());
+	std::vector<matocsserventry *> all_servers;
+	std::vector<ChunkPartType> all_parts;
+	ChunkCopiesCalculator calc(c->getGoal());
+
+	uint32_t destination_version = matocsserv_get_version(destination_server);
+
+	assert(destination_version >= getMinChunkserverVersion(c, part_to_recover));
 
 	for (slist *s = c->slisthead; s; s = s->next) {
-		if (s->is_valid() && !s->is_busy()) {
-			if (matocsserv_get_version(s->ptr) >= kFirstXorVersion) {
-				xor_capable_servers.push_back(s->ptr);
-				xor_capable_parts.push_back(s->chunkType);
-				xor_capable_calc.addPart(s->chunkType,
-				                         matocsserv_get_label(s->ptr));
-			}
-			if (slice_traits::isStandard(s->chunkType)) {
-				standard_servers.push_back(s->ptr);
-			}
+		if (!s->is_valid() || s->is_busy() || matocsserv_replication_read_counter(s->ptr) >= MaxReadRepl) {
+			continue;
 		}
+
+		if (slice_traits::isStandard(s->chunkType)) {
+			standard_servers.push_back(s->ptr);
+		}
+
+		if (destination_version >= kFirstXorVersion && destination_version < kFirstECVersion
+			&& slice_traits::isXor(part_to_recover) && matocsserv_get_version(s->ptr) < kFirstXorVersion) {
+			continue;
+		}
+
+		if (destination_version < kFirstXorVersion && !slice_traits::isStandard(s->chunkType)) {
+			continue;
+		}
+
+		all_servers.push_back(s->ptr);
+		all_parts.push_back(s->chunkType);
+		calc.addPart(s->chunkType, matocsserv_get_label(s->ptr));
 	}
 
-	// we calculate only chunk state here
-	// target optimization is not required
-	xor_capable_calc.evalState();
-
-	if (xor_capable_calc.isRecoveryPossible() &&
-	    matocsserv_get_version(destinationServer) >= kFirstXorVersion) {
-		// new replication possible - use it
-		matocsserv_send_liz_replicatechunk(destinationServer, c->chunkid, c->version,
-		                                   part_to_recover, xor_capable_servers,
-		                                   xor_capable_parts);
-	} else if (slice_traits::isStandard(part_to_recover) && !standard_servers.empty()) {
-		// fall back to legacy replication
-		matocsserv_send_replicatechunk(
-		        destinationServer, c->chunkid, c->version,
-		        standard_servers[rnd_ranged<uint32_t>(standard_servers.size())]);
-	} else {
-		// no replication possible
+	calc.evalState();
+	if (!calc.isRecoveryPossible()) {
 		return false;
 	}
+
+	if (destination_version >= kFirstECVersion ||
+	    (destination_version >= kFirstXorVersion && slice_traits::isXor(part_to_recover))) {
+		matocsserv_send_liz_replicatechunk(destination_server, c->chunkid, c->version,
+		                                   part_to_recover, all_servers,
+		                                   all_parts);
+		stats_replications++;
+		c->needverincrease = 1;
+		return true;
+	}
+
+	// fall back to legacy replication
+	assert(slice_traits::isStandard(part_to_recover));
+
+	if (standard_servers.empty()) {
+		return false;
+	}
+
+	matocsserv_send_replicatechunk(destination_server, c->chunkid, c->version,
+	                               standard_servers[rnd_ranged<uint32_t>(standard_servers.size())]);
+
 	stats_replications++;
 	c->needverincrease = 1;
 	return true;
@@ -2131,6 +2166,8 @@ bool ChunkWorker::replicateChunkPart(chunk *c, Goal::Slice::Type slice_type, int
 
 	expected_copies = Goal::Slice::countLabels(calc.getTarget()[slice_type][slice_part]);
 
+	uint32_t min_chunkserver_version = getMinChunkserverVersion(c, ChunkPartType(slice_type, slice_part));
+
 	for (const auto &label_and_count : replicate_labels) {
 		tried_to_replicate = true;
 
@@ -2151,8 +2188,9 @@ bool ChunkWorker::replicateChunkPart(chunk *c, Goal::Slice::Type slice_type, int
 
 		// Get a list of possible destination servers
 		int total_matching, returned_matching, temporarily_unavailable;
-		matocsserv_getservers_lessrepl(label_and_count.first, MaxWriteRepl, servers,
-		                               total_matching, returned_matching, temporarily_unavailable);
+		matocsserv_getservers_lessrepl(label_and_count.first, min_chunkserver_version, MaxWriteRepl,
+		                               servers, total_matching, returned_matching,
+		                               temporarily_unavailable);
 
 		// Find a destination server for replication -- the first one without a copy of 'c'
 		matocsserventry *destination = nullptr;
@@ -2292,7 +2330,7 @@ bool ChunkWorker::rebalanceChunkParts(chunk *c, ChunkCopiesCalculator &calc, boo
 	// There are at least two servers with a disk usage difference grater than
 	// AcceptableDifference, so it's worth checking.
 	for (slist *s = c->slisthead; s != nullptr; s = s->next) {
-		if (!s->is_valid() || matocsserv_replication_read_counter(s->ptr) >= MaxReadRepl) {
+		if (!s->is_valid()) {
 			continue;
 		}
 
@@ -2316,6 +2354,8 @@ bool ChunkWorker::rebalanceChunkParts(chunk *c, ChunkCopiesCalculator &calc, boo
 		                                          s->chunkType.getSlicePart(),
 		                                          current_copy_label));
 
+		uint32_t min_chunkserver_version = getMinChunkserverVersion(c, s->chunkType);
+
 		const ServersWithUsage &sorted_servers =
 		        multi_label_rebalance ? sortedServers_
 		                              : labeledSortedServers_[current_copy_label];
@@ -2325,16 +2365,13 @@ bool ChunkWorker::rebalanceChunkParts(chunk *c, ChunkCopiesCalculator &calc, boo
 				break;  // No more suitable destination servers (next servers have
 				        // higher usage)
 			}
-			if (slice_traits::isXor(s->chunkType) &&
-			    matocsserv_get_version(empty_server.server) < kFirstXorVersion) {
-				continue;  // We can't place xor chunks on old servers
+			if (matocsserv_get_version(empty_server.server) < min_chunkserver_version) {
+				continue;
 			}
-			if (chunkPresentOnServer(c, s->chunkType.getSliceType(),
-			                         empty_server.server)) {
+			if (chunkPresentOnServer(c, s->chunkType.getSliceType(), empty_server.server)) {
 				continue;  // A copy is already here
 			}
-			if (matocsserv_replication_write_counter(empty_server.server) >=
-			    MaxWriteRepl) {
+			if (matocsserv_replication_write_counter(empty_server.server) >= MaxWriteRepl) {
 				continue;  // We can't create a new copy here
 			}
 			if (tryReplication(c, s->chunkType, empty_server.server)) {
