@@ -33,6 +33,7 @@
 #include "master/filesystem_freenode.h"
 #include "master/filesystem_metadata.h"
 #include "master/filesystem_operations.h"
+#include "master/filesystem_quota.h"
 #include "master/fs_context.h"
 
 #ifndef NDEBUG
@@ -93,12 +94,12 @@ static uint64_t file_size(fsnode *node, uint32_t nonzero_chunks) {
 }
 
 #ifndef METARESTORE
-// compute the disk space cost of all parts of a xor chunk of given size
-static uint32_t xor_chunk_realsize(uint32_t blocks, uint32_t level) {
-	const uint32_t stripes = (blocks + level - 1) / level;
-	uint32_t size = blocks * MFSBLOCKSIZE;  // file data
-	size += stripes * MFSBLOCKSIZE;         // parity data
-	size += 4096 * (level + 1);             // headers of data and parity parts
+// compute the disk space cost of all parts of a xor/ec chunk of given size
+static uint32_t ec_chunk_realsize(uint32_t blocks, uint32_t data_part_count, uint32_t parity_part_count) {
+	const uint32_t stripes = (blocks + data_part_count - 1) / data_part_count;
+	uint32_t size = blocks * MFSBLOCKSIZE;                 // file data
+	size += parity_part_count * stripes * MFSBLOCKSIZE;     // parity data
+	size += 4096 * (data_part_count + parity_part_count);  // headers of data and parity parts
 	return size;
 }
 #endif
@@ -118,19 +119,22 @@ static uint64_t file_realsize(fsnode *node, uint32_t nonzero_chunks, uint64_t fi
 	for (const auto &slice : goal) {
 		if (slice_traits::isStandard(slice) || slice_traits::isTape(slice)) {
 			full_size += file_size * slice.getExpectedCopies();
-		} else if (slice_traits::isXor(slice)) {
-			int level = slice_traits::xors::getXorLevel(slice);
-			uint32_t full_chunk_realsize = xor_chunk_realsize(MFSBLOCKSINCHUNK, level);
+		} else if (slice_traits::isXor(slice) || slice_traits::isEC(slice)) {
+			int data_part_count = slice_traits::getNumberOfDataParts(slice);
+			int parity_part_count = slice_traits::getNumberOfParityParts(slice);
+
+			uint32_t full_chunk_realsize =
+			    ec_chunk_realsize(MFSBLOCKSINCHUNK, data_part_count, parity_part_count);
 			uint64_t size = (uint64_t)nonzero_chunks * full_chunk_realsize;
 			if (last_chunk_nonempty(node)) {
 				size -= full_chunk_realsize;
-				size += xor_chunk_realsize(last_chunk_blocks(node), level);
+				size +=
+				    ec_chunk_realsize(last_chunk_blocks(node), data_part_count, parity_part_count);
 			}
 			full_size += size;
 		} else {
-			syslog(LOG_ERR,
-			       "file_realsize: inode %" PRIu32 " has unknown goal 0x%" PRIx8,
-			       node->id, node->goal);
+			syslog(LOG_ERR, "file_realsize: inode %" PRIu32 " has unknown goal 0x%" PRIx8, node->id,
+			       node->goal);
 			return 0;
 		}
 	}
@@ -271,33 +275,6 @@ bool fsnodes_isancestor_or_node_reserved_or_trash(fsnode *f, fsnode *p) {
 	return fsnodes_isancestor(f, p);
 }
 
-// quota
-
-bool fsnodes_inode_quota_exceeded(uint32_t uid, uint32_t gid) {
-	return gMetadata->gQuotaDatabase.isExceeded(QuotaRigor::kHard, QuotaResource::kInodes, uid,
-	                                            gid);
-}
-
-bool fsnodes_size_quota_exceeded(uint32_t uid, uint32_t gid) {
-	return gMetadata->gQuotaDatabase.isExceeded(QuotaRigor::kHard, QuotaResource::kSize, uid,
-	                                            gid);
-}
-
-void fsnodes_quota_register_inode(fsnode *node) {
-	gMetadata->gQuotaDatabase.changeUsage(QuotaResource::kInodes, node->uid, node->gid, +1);
-}
-
-static void fsnodes_quota_unregister_inode(fsnode *node) {
-	gMetadata->gQuotaDatabase.changeUsage(QuotaResource::kInodes, node->uid, node->gid, -1);
-}
-
-void fsnodes_quota_update_size(fsnode *node, int64_t delta) {
-	if (delta != 0) {
-		gMetadata->gQuotaDatabase.changeUsage(QuotaResource::kSize, node->uid, node->gid,
-		                                      delta);
-	}
-}
-
 // stats
 
 void fsnodes_get_stats(fsnode *node, statsrecord *sr) {
@@ -678,9 +655,9 @@ fsnode *fsnodes_create_node(uint32_t ts, fsnode *node, uint16_t nleng, const uin
 	p->checksum = 0;
 	fsnodes_update_checksum(p);
 	fsnodes_link(ts, node, p, nleng, name);
-	fsnodes_quota_register_inode(p);
+	fsnodes_quota_update(p, {{QuotaResource::kInodes, +1}});
 	if (type == TYPE_FILE) {
-		fsnodes_quota_update_size(p, +fsnodes_get_size(p));
+		fsnodes_quota_update(p, {{QuotaResource::kSize, +fsnodes_get_size(p)}});
 	}
 	return p;
 }
@@ -1014,7 +991,7 @@ uint8_t fsnodes_appendchunks(uint32_t ts, fsnode *dstobj, fsnode *srcobj) {
 	}
 	dstobj->data.fdata.length = length;
 	fsnodes_get_stats(dstobj, &nsr);
-	fsnodes_quota_update_size(dstobj, nsr.size - psr.size);
+	fsnodes_quota_update(dstobj, {{QuotaResource::kSize, nsr.size - psr.size}});
 	for (fsedge *e = dstobj->parents; e; e = e->nextparent) {
 		fsnodes_add_sub_stats(e->parent, &nsr, &psr);
 	}
@@ -1092,7 +1069,7 @@ void fsnodes_setlength(fsnode *obj, uint64_t length) {
 		}
 	}
 	fsnodes_get_stats(obj, &nsr);
-	fsnodes_quota_update_size(obj, nsr.size - psr.size);
+	fsnodes_quota_update(obj, {{QuotaResource::kSize, nsr.size - psr.size}});
 	for (fsedge *e = obj->parents; e; e = e->nextparent) {
 		fsnodes_add_sub_stats(e->parent, &nsr, &psr);
 	}
@@ -1101,16 +1078,16 @@ void fsnodes_setlength(fsnode *obj, uint64_t length) {
 
 void fsnodes_change_uid_gid(fsnode *p, uint32_t uid, uint32_t gid) {
 	int64_t size = 0;
-	fsnodes_quota_unregister_inode(p);
+	fsnodes_quota_update(p, {{QuotaResource::kInodes, -1}});
 	if (p->type == TYPE_FILE || p->type == TYPE_TRASH || p->type == TYPE_RESERVED) {
 		size = fsnodes_get_size(p);
-		fsnodes_quota_update_size(p, -size);
+		fsnodes_quota_update(p, {{QuotaResource::kSize, -size}});
 	}
 	p->uid = uid;
 	p->gid = gid;
-	fsnodes_quota_register_inode(p);
+	fsnodes_quota_update(p, {{QuotaResource::kInodes, +1}});
 	if (p->type == TYPE_FILE || p->type == TYPE_TRASH || p->type == TYPE_RESERVED) {
-		fsnodes_quota_update_size(p, +size);
+		fsnodes_quota_update(p, {{QuotaResource::kSize, +size}});
 	}
 }
 
@@ -1143,7 +1120,7 @@ static inline void fsnodes_remove_node(uint32_t ts, fsnode *toremove) {
 	    toremove->type == TYPE_RESERVED) {
 		uint32_t i;
 		uint64_t chunkid;
-		fsnodes_quota_update_size(toremove, -fsnodes_get_size(toremove));
+		fsnodes_quota_update(toremove, {{QuotaResource::kSize, -fsnodes_get_size(toremove)}});
 		gMetadata->filenodes--;
 		for (i = 0; i < toremove->data.fdata.chunks; i++) {
 			chunkid = toremove->data.fdata.chunktab[i];
@@ -1159,7 +1136,8 @@ static inline void fsnodes_remove_node(uint32_t ts, fsnode *toremove) {
 	}
 	gMetadata->inode_pool.release(toremove->id, ts, true);
 	xattr_removeinode(toremove->id);
-	fsnodes_quota_unregister_inode(toremove);
+	fsnodes_quota_update(toremove, {{QuotaResource::kInodes, -1}});
+	fsnodes_quota_remove(QuotaOwnerType::kInode, toremove->id);
 #ifndef METARESTORE
 	dcm_modify(toremove->id, 0);
 #endif
@@ -1185,7 +1163,7 @@ void fsnodes_unlink(uint32_t ts, fsedge *e) {
 		if (child->type == TYPE_FILE) {
 			if (child->trashtime > 0) {
 				child->type = TYPE_TRASH;
-				child->ctime = ts;
+				child->atime = ts;
 				fsnodes_update_checksum(child);
 				e = new fsedge;
 				e->nleng = pleng;
