@@ -1,8 +1,7 @@
 /*
-   Copyright 2005-2010 Jakub Kruszona-Zawadzki, Gemius SA, 2013-2014 EditShare, 2013-2015
-   Skytechnology sp. z o.o..
+   Copyright 2013-2015 Skytechnology sp. z o.o..
 
-   This file was part of MooseFS and is part of LizardFS.
+   This file is part of LizardFS.
 
    LizardFS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -30,6 +29,7 @@
 #include "master/filesystem_checksum.h"
 #include "master/filesystem_checksum_updater.h"
 #include "master/filesystem_node.h"
+#include "master/filesystem_quota.h"
 #include "master/fs_context.h"
 #include "master/locks.h"
 #include "master/matocsserv.h"
@@ -376,6 +376,7 @@ void fs_statfs(uint32_t rootinode, uint8_t sesflags, uint64_t *totalspace, uint6
 		*inodes = 0;
 	} else {
 		matocsserv_getspace(totalspace, availspace);
+		fsnodes_quota_adjust_space(rn, *totalspace, *availspace);
 		fsnodes_get_stats(rn, &sr);
 		*inodes = sr.inodes;
 		if (sr.realsize + *availspace < *totalspace) {
@@ -615,7 +616,7 @@ uint8_t fs_try_setlength(uint32_t rootinode, uint8_t sesflags, uint32_t inode, u
 				status = chunk_multi_truncate(
 				        ochunkid, lockId, (length & MFSCHUNKMASK), p->goal,
 				        denyTruncatingParity,
-				        fsnodes_size_quota_exceeded(p->uid, p->gid), &nchunkid);
+				        fsnodes_quota_exceeded(p, {{QuotaResource::kSize, 1}}), &nchunkid);
 				if (status != LIZARDFS_STATUS_OK) {
 					return status;
 				}
@@ -1010,8 +1011,9 @@ uint8_t fs_symlink(const FsContext &context, uint32_t parent, uint16_t nleng, co
 	if (fsnodes_nameisused(wd, nleng, name)) {
 		return LIZARDFS_ERROR_EEXIST;
 	}
-	if ((context.isPersonalityMaster()) &&
-	    fsnodes_inode_quota_exceeded(context.uid(), context.gid())) {
+	if (context.isPersonalityMaster() &&
+	    (fsnodes_quota_exceeded_ug(context.uid(), context.gid(), {{QuotaResource::kInodes, 1}}) ||
+	     fsnodes_quota_exceeded_dir(wd, {{QuotaResource::kInodes, 1}}))) {
 		return LIZARDFS_ERROR_QUOTA;
 	}
 	uint8_t *newpath = (uint8_t *)malloc(pleng);
@@ -1101,7 +1103,8 @@ uint8_t fs_mknod(uint32_t rootinode, uint8_t sesflags, uint32_t parent, uint16_t
 	if (fsnodes_nameisused(wd, nleng, name)) {
 		return LIZARDFS_ERROR_EEXIST;
 	}
-	if (fsnodes_inode_quota_exceeded(uid, gid)) {
+	if (fsnodes_quota_exceeded_ug(uid, gid, {{QuotaResource::kInodes, 1}}) ||
+	    fsnodes_quota_exceeded_dir(wd, {{QuotaResource::kInodes, 1}})) {
 		return LIZARDFS_ERROR_QUOTA;
 	}
 	p = fsnodes_create_node(ts, wd, nleng, name, type, mode, umask, uid, gid, 0,
@@ -1167,7 +1170,8 @@ uint8_t fs_mkdir(uint32_t rootinode, uint8_t sesflags, uint32_t parent, uint16_t
 	if (fsnodes_nameisused(wd, nleng, name)) {
 		return LIZARDFS_ERROR_EEXIST;
 	}
-	if (fsnodes_inode_quota_exceeded(uid, gid)) {
+	if (fsnodes_quota_exceeded_ug(uid, gid, {{QuotaResource::kInodes, 1}}) ||
+	    fsnodes_quota_exceeded_dir(wd, {{QuotaResource::kInodes, 1}})) {
 		return LIZARDFS_ERROR_QUOTA;
 	}
 	p = fsnodes_create_node(ts, wd, nleng, name, TYPE_DIRECTORY, mode, umask, uid, gid,
@@ -1400,11 +1404,18 @@ uint8_t fs_rename(const FsContext &context, uint32_t parent_src, uint16_t nleng_
 	} else {
 		*inode = node->id;
 	}
+
+	std::array<int64_t, 2> quota_delta = {{1, 1}};
 	if (se->child->type == TYPE_DIRECTORY) {
 		if (fsnodes_isancestor(se->child, dwd)) {
 			return LIZARDFS_ERROR_EINVAL;
 		}
+		statsrecord &stats = *se->child->data.ddata.stats;
+		quota_delta = {{(int64_t)stats.inodes, (int64_t)stats.size}};
+	} else if (se->child->type == TYPE_FILE) {
+		quota_delta[(int)QuotaResource::kSize] = fsnodes_get_size(se->child);
 	}
+
 	if (fsnodes_namecheck(nleng_dst, name_dst) < 0) {
 		return LIZARDFS_ERROR_EINVAL;
 	}
@@ -1417,6 +1428,26 @@ uint8_t fs_rename(const FsContext &context, uint32_t parent_src, uint16_t nleng_
 		    !fsnodes_sticky_access(dwd, de->child, context.uid())) {
 			return LIZARDFS_ERROR_EPERM;
 		}
+		if (de->child->type == TYPE_DIRECTORY) {
+			statsrecord &stats = *de->child->data.ddata.stats;
+			quota_delta[(int)QuotaResource::kInodes] -= stats.inodes;
+			quota_delta[(int)QuotaResource::kSize] -= stats.size;
+		} else if (de->child->type == TYPE_FILE) {
+			quota_delta[(int)QuotaResource::kInodes] -= 1;
+			quota_delta[(int)QuotaResource::kSize] -= fsnodes_get_size(de->child);
+		} else {
+			quota_delta[(int)QuotaResource::kInodes] -= 1;
+			quota_delta[(int)QuotaResource::kSize] -= 1;
+		}
+	}
+
+	if (fsnodes_quota_exceeded_dir(
+	        dwd, swd, {{QuotaResource::kInodes, quota_delta[(int)QuotaResource::kInodes]},
+	                   {QuotaResource::kSize, quota_delta[(int)QuotaResource::kSize]}})) {
+		return LIZARDFS_ERROR_QUOTA;
+	}
+
+	if (de) {
 		fsnodes_unlink(context.ts(), de);
 	}
 	fsnodes_remove_edge(context.ts(), se);
@@ -1504,7 +1535,7 @@ uint8_t fs_append(const FsContext &context, uint32_t inode, uint32_t inode_src)
 	if (status != LIZARDFS_STATUS_OK) {
 		return status;
 	}
-	if (context.isPersonalityMaster() && fsnodes_size_quota_exceeded(p->uid, p->gid)) {
+	if (context.isPersonalityMaster() && fsnodes_quota_exceeded(p, {{QuotaResource::kSize, 1}})) {
 		return LIZARDFS_ERROR_QUOTA;
 	}
 	status = fsnodes_appendchunks(context.ts(), p, sp);
@@ -2110,7 +2141,7 @@ uint8_t fs_writechunk(const FsContext &context, uint32_t inode, uint32_t indx, b
 	}
 #endif
 
-	const bool quota_exceeded = fsnodes_size_quota_exceeded(p->uid, p->gid);
+	const bool quota_exceeded = fsnodes_quota_exceeded(p, {{QuotaResource::kSize, 1}});
 	statsrecord psr;
 	fsnodes_get_stats(p, &psr);
 
@@ -2171,7 +2202,7 @@ uint8_t fs_writechunk(const FsContext &context, uint32_t inode, uint32_t indx, b
 	for (fsedge *e = p->parents; e; e = e->nextparent) {
 		fsnodes_add_sub_stats(e->parent, &nsr, &psr);
 	}
-	fsnodes_quota_update_size(p, nsr.size - psr.size);
+	fsnodes_quota_update(p, {{QuotaResource::kSize, nsr.size - psr.size}});
 	if (length) {
 		*length = p->data.fdata.length;
 	}
@@ -2300,7 +2331,7 @@ uint8_t fs_repair(uint32_t rootinode, uint8_t sesflags, uint32_t inode, uint32_t
 	for (e = p->parents; e; e = e->nextparent) {
 		fsnodes_add_sub_stats(e->parent, &nsr, &psr);
 	}
-	fsnodes_quota_update_size(p, nsr.size - psr.size);
+	fsnodes_quota_update(p, {{QuotaResource::kSize, nsr.size - psr.size}});
 	fsnodes_update_checksum(p);
 	return LIZARDFS_STATUS_OK;
 }
@@ -2338,7 +2369,7 @@ uint8_t fs_apply_repair(uint32_t ts, uint32_t inode, uint32_t indx, uint32_t nve
 	for (fsedge *e = p->parents; e; e = e->nextparent) {
 		fsnodes_add_sub_stats(e->parent, &nsr, &psr);
 	}
-	fsnodes_quota_update_size(p, nsr.size - psr.size);
+	fsnodes_quota_update(p, {{QuotaResource::kSize, nsr.size - psr.size}});
 	gMetadata->metaversion++;
 	p->mtime = p->ctime = ts;
 	fsnodes_update_checksum(p);
@@ -2865,89 +2896,6 @@ uint8_t fs_apply_setacl(uint32_t ts, uint32_t inode, char aclType, const char *a
 	return status;
 }
 
-#ifndef METARESTORE
-uint8_t fs_quota_get_all(uint8_t sesflags, uint32_t uid,
-			std::vector<QuotaOwnerAndLimits> &results) {
-	if (uid != 0 && !(sesflags & SESFLAG_ALLCANCHANGEQUOTA)) {
-		return LIZARDFS_ERROR_EPERM;
-	}
-	results = gMetadata->gQuotaDatabase.getAll();
-	return LIZARDFS_STATUS_OK;
-}
-
-uint8_t fs_quota_get(uint8_t sesflags, uint32_t uid, uint32_t gid,
-		const std::vector<QuotaOwner> &owners,
-		std::vector<QuotaOwnerAndLimits> &results) {
-	std::vector<QuotaOwnerAndLimits> tmp;
-	for (const QuotaOwner &owner : owners) {
-		if (uid != 0 && !(sesflags & SESFLAG_ALLCANCHANGEQUOTA)) {
-			switch (owner.ownerType) {
-			case QuotaOwnerType::kUser:
-				if (uid != owner.ownerId) {
-					return LIZARDFS_ERROR_EPERM;
-				}
-				break;
-			case QuotaOwnerType::kGroup:
-				if (gid != owner.ownerId && !(sesflags & SESFLAG_IGNOREGID)) {
-					return LIZARDFS_ERROR_EPERM;
-				}
-				break;
-			default:
-				return LIZARDFS_ERROR_EINVAL;
-			}
-		}
-		const QuotaLimits *result =
-		        gMetadata->gQuotaDatabase.get(owner.ownerType, owner.ownerId);
-		if (result) {
-			tmp.emplace_back(owner, *result);
-		}
-	}
-	results.swap(tmp);
-	return LIZARDFS_STATUS_OK;
-}
-
-uint8_t fs_quota_set(uint8_t sesflags, uint32_t uid, const std::vector<QuotaEntry> &entries) {
-	uint32_t ts = main_time();
-	ChecksumUpdater cu(ts);
-	if (sesflags & SESFLAG_READONLY) {
-		return LIZARDFS_ERROR_EROFS;
-	}
-	if (uid != 0 && !(sesflags & SESFLAG_ALLCANCHANGEQUOTA)) {
-		return LIZARDFS_ERROR_EPERM;
-	}
-	for (const QuotaEntry &entry : entries) {
-		const QuotaOwner &owner = entry.entryKey.owner;
-		gMetadata->gQuotaDatabase.set(entry.entryKey.rigor, entry.entryKey.resource,
-		                              owner.ownerType, owner.ownerId, entry.limit);
-		fs_changelog(ts, "SETQUOTA(%c,%c,%c,%" PRIu32 ",%" PRIu64 ")",
-		             (entry.entryKey.rigor == QuotaRigor::kSoft) ? 'S' : 'H',
-		             (entry.entryKey.resource == QuotaResource::kSize) ? 'S' : 'I',
-		             (owner.ownerType == QuotaOwnerType::kUser) ? 'U' : 'G',
-		             uint32_t{owner.ownerId}, uint64_t{entry.limit});
-	}
-	return LIZARDFS_STATUS_OK;
-}
-#endif
-
-uint8_t fs_apply_setquota(char rigor, char resource, char ownerType, uint32_t ownerId,
-			uint64_t limit) {
-	QuotaRigor quotaRigor = QuotaRigor::kSoft;
-	QuotaResource quotaResource = QuotaResource::kSize;
-	QuotaOwnerType quotaOwnerType = QuotaOwnerType::kUser;
-	bool valid = true;
-	valid &= decodeChar("SH", {QuotaRigor::kSoft, QuotaRigor::kHard}, rigor, quotaRigor);
-	valid &= decodeChar("SI", {QuotaResource::kSize, QuotaResource::kInodes}, resource,
-	                    quotaResource);
-	valid &= decodeChar("UG", {QuotaOwnerType::kUser, QuotaOwnerType::kGroup}, ownerType,
-	                    quotaOwnerType);
-	if (!valid) {
-		return LIZARDFS_ERROR_EINVAL;
-	}
-	gMetadata->metaversion++;
-	gMetadata->gQuotaDatabase.set(quotaRigor, quotaResource, quotaOwnerType, ownerId, limit);
-	return LIZARDFS_STATUS_OK;
-}
-
 #ifndef METARESTORE
 uint32_t fs_getdirpath_size(uint32_t inode) {
 	fsnode *node;
