@@ -39,6 +39,7 @@
 #include "master/filesystem_freenode.h"
 #include "master/filesystem_operations.h"
 #include "master/filesystem_checksum.h"
+#include "master/filesystem_quota.h"
 #include "master/locks.h"
 #include "master/matoclserv.h"
 #include "master/matomlserv.h"
@@ -799,7 +800,7 @@ int fs_loadnode(FILE *fd) {
 #endif
 			sessionids--;
 		}
-		fsnodes_quota_update_size(p, +fsnodes_get_size(p));
+		fsnodes_quota_update(p, {{QuotaResource::kSize, +fsnodes_get_size(p)}});
 	}
 	p->parents = NULL;
 	nodepos = NODEHASHPOS(p->id);
@@ -813,7 +814,7 @@ int fs_loadnode(FILE *fd) {
 	if (type == TYPE_FILE || type == TYPE_TRASH || type == TYPE_RESERVED) {
 		gMetadata->filenodes++;
 	}
-	fsnodes_quota_register_inode(p);
+	fsnodes_quota_update(p, {{QuotaResource::kInodes, +1}});
 	return 0;
 }
 
@@ -864,7 +865,7 @@ static void fs_storeacls(FILE *fd) {
 }
 
 static void fs_storequotas(FILE *fd) {
-	const std::vector<QuotaEntry> &entries = gMetadata->gQuotaDatabase.getEntries();
+	const std::vector<QuotaEntry> &entries = gMetadata->quota_database.getEntries();
 	fs_store_generic(fd, entries);
 }
 
@@ -958,10 +959,11 @@ static int fs_loadquotas(FILE *fd, int ignoreflag) {
 		std::vector<QuotaEntry> entries;
 		fs_load_generic(fd, entries);
 		for (const auto &entry : entries) {
-			gMetadata->gQuotaDatabase.set(entry.entryKey.rigor, entry.entryKey.resource,
-			                              entry.entryKey.owner.ownerType,
-			                              entry.entryKey.owner.ownerId, entry.limit);
+			gMetadata->quota_database.set(entry.entryKey.owner.ownerType,
+			                              entry.entryKey.owner.ownerId, entry.entryKey.rigor,
+			                              entry.entryKey.resource, entry.limit);
 		}
+		gMetadata->quota_checksum = gMetadata->quota_database.checksum();
 	} catch (Exception &ex) {
 		lzfs_pretty_syslog(LOG_ERR, "loading quotas: %s", ex.what());
 		if (!ignoreflag || ex.status() != LIZARDFS_STATUS_OK) {
@@ -1019,7 +1021,7 @@ void fs_storefree(FILE *fd) {
 	}
 }
 
-int fs_loadfree(FILE *fd) {
+int fs_loadfree(FILE *fd, uint64_t section_size = 0) {
 	uint8_t rbuff[8 * 1024];
 	const uint8_t *ptr;
 	uint32_t l, t;
@@ -1031,6 +1033,11 @@ int fs_loadfree(FILE *fd) {
 	ptr = rbuff;
 	t = get32bit(&ptr);
 
+	if (section_size && t != (section_size - 4) / 8) {
+		lzfs_pretty_errlog(LOG_INFO, "loading free nodes: section size doesn't match number of free nodes");
+		t = (section_size - 4) / 8;
+	}
+
 	l = 0;
 	while (t > 0) {
 		if (l == 0) {
@@ -1275,7 +1282,7 @@ int fs_load(FILE *fd, int ignoreflag, uint8_t fver) {
 				        LOG_INFO,
 				        "loading deletion timestamps from the metadata file");
 				fflush(stderr);
-				if (fs_loadfree(fd) < 0) {
+				if (fs_loadfree(fd, sleng) < 0) {
 #ifndef METARESTORE
 					lzfs_pretty_syslog(LOG_ERR,
 					                   "error reading metadata (free)");
@@ -1418,7 +1425,7 @@ void fs_new(void) {
 	gMetadata->dirnodes = 1;
 	gMetadata->filenodes = 0;
 	fs_checksum(ChecksumMode::kForceRecalculate);
-	fsnodes_quota_register_inode(gMetadata->root);
+	fsnodes_quota_update(gMetadata->root, {{QuotaResource::kInodes, +1}});
 }
 #endif
 
