@@ -40,6 +40,7 @@
 #include "common/goal.h"
 #include "common/hashfn.h"
 #include "common/lizardfs_version.h"
+#include "common/loop_watchdog.h"
 #include "common/main.h"
 #include "common/massert.h"
 #include "common/mfserr.h"
@@ -465,14 +466,27 @@ std::vector<std::pair<matocsserventry *, ChunkPartType>> matocsserv_getservers_f
 		std::array<int, Goal::Slice::kMaxPartsCount> shuffle;
 
 		std::iota(shuffle.begin(), shuffle.begin() + slice.size(), 0);
-		std::random_shuffle(shuffle.begin(), shuffle.begin() + slice.size());
+		// Move xor parity to the end of a list
+		if (slice_traits::isXor(slice)) {
+			std::swap(shuffle[0], shuffle[slice.size() - 1]);
+		}
+		// Shuffle data before parity to prioritize data parts before parity in partial writes
+		int data_count = slice_traits::getNumberOfDataParts(slice);
+		assert(std::all_of(shuffle.begin(), shuffle.begin() + data_count,
+				[&slice](int i){
+			return slice_traits::isDataPart(ChunkPartType(slice.getType(), i));
+		}));
+		std::random_shuffle(shuffle.begin(), shuffle.begin() + data_count);
+		std::random_shuffle(shuffle.begin() + data_count, shuffle.begin() + slice.size());
+
+		uint32_t min_version = std::max({
+			slice_traits::isXor(slice) ? kFirstXorVersion : 0,
+			slice_traits::isEC(slice) ? kFirstECVersion : 0,
+			min_server_version
+		});
 
 		int count_full_parts = 0;
 		for (int i = 0; i < slice.size(); ++i) {
-			uint32_t min_version = std::max(
-				slice_traits::isXor(slice) ? kFirstXorVersion : 0,
-				min_server_version
-			);
 			servers = getter.chooseServersForLabels(
 			        history[goal_id], slice[shuffle[i]],
 			        min_version, used_servers);
@@ -486,8 +500,7 @@ std::vector<std::pair<matocsserventry *, ChunkPartType>> matocsserv_getservers_f
 			}
 		}
 
-		if (slice_traits::isXor(slice) &&
-		    count_full_parts < slice_traits::xors::getXorLevel(slice)) {
+		if (count_full_parts < slice_traits::getNumberOfDataParts(slice.getType())) {
 			continue; // do not create any parts if servers are missing
 		}
 
@@ -497,9 +510,9 @@ std::vector<std::pair<matocsserventry *, ChunkPartType>> matocsserv_getservers_f
 	return ret;
 }
 
-void matocsserv_getservers_lessrepl(const MediaLabel &label, uint16_t replication_write_limit,
-		std::vector<matocsserventry*> &servers, int &total_matching, int &returned_matching,
-		int &temporarily_unavailable) {
+void matocsserv_getservers_lessrepl(const MediaLabel &label, uint32_t min_chunkserver_version,
+		uint16_t replication_write_limit, std::vector<matocsserventry *> &servers,
+		int &total_matching, int &returned_matching, int &temporarily_unavailable) {
 	total_matching = 0;
 	returned_matching = 0;
 	temporarily_unavailable = 0;
@@ -512,6 +525,9 @@ void matocsserv_getservers_lessrepl(const MediaLabel &label, uint16_t replicatio
 			// Skip full and disconnected chunkservers
 			continue;
 		}
+		if (eptr->version < min_chunkserver_version) {
+			continue;
+		}
 		bool matchesRequestedLabel = false;
 		if (label != MediaLabel::kWildcard && eptr->label == label) {
 			++total_matching;
@@ -655,6 +671,10 @@ int matocsserv_send_createchunk(matocsserventry *eptr, uint64_t chunkId, ChunkPa
 			sassert(slice_traits::isStandard(chunkType));
 			serializeMooseFsPacket(eptr->outputPackets.back().packet, MATOCS_CREATE, chunkId,
 					chunkVersion);
+		} else if (eptr->version < kFirstECVersion) {
+			sassert((int)chunkType.getSliceType() < Goal::Slice::Type::kECFirst);
+			matocs::createChunk::serialize(eptr->outputPackets.back().packet, chunkId, (legacy::ChunkPartType)chunkType,
+					chunkVersion);
 		} else {
 			matocs::createChunk::serialize(eptr->outputPackets.back().packet, chunkId, chunkType,
 					chunkVersion);
@@ -672,7 +692,15 @@ void matocsserv_got_createchunk_status(matocsserventry *eptr, const std::vector<
 		deserializeAllMooseFsPacketDataNoHeader(data, chunkId, status);
 	}
 	else {
-		cstoma::createChunk::deserialize(data, chunkId, chunkType, status);
+		PacketVersion v;
+		deserializePacketVersionNoHeader(data, v);
+		if (v == cstoma::createChunk::kECChunks) {
+			cstoma::createChunk::deserialize(data, chunkId, chunkType, status);
+		} else {
+			legacy::ChunkPartType legacy_type;
+			cstoma::createChunk::deserialize(data, chunkId, legacy_type, status);
+			chunkType = legacy_type;
+		}
 	}
 	chunk_got_create_status(eptr, chunkId, chunkType, status);
 	if (status != 0) {
@@ -691,7 +719,11 @@ int matocsserv_send_deletechunk(matocsserventry *eptr, uint64_t chunkId, uint32_
 			serializeMooseFsPacket(eptr->outputPackets.back().packet, MATOCS_DELETE,
 					chunkId, chunkVersion);
 		}
-		else {
+		else if (eptr->version < kFirstECVersion) {
+			sassert((int)chunkType.getSliceType() < Goal::Slice::Type::kECFirst);
+			matocs::deleteChunk::serialize(eptr->outputPackets.back().packet,
+					chunkId, (legacy::ChunkPartType)chunkType, chunkVersion);
+		} else {
 			matocs::deleteChunk::serialize(eptr->outputPackets.back().packet,
 					chunkId, chunkType, chunkVersion);
 		}
@@ -708,7 +740,15 @@ void matocsserv_got_deletechunk_status(matocsserventry *eptr, const std::vector<
 	if (eptr->version < kFirstXorVersion) {
 		deserializeAllMooseFsPacketDataNoHeader(data, chunkId, status);
 	} else {
-		cstoma::deleteChunk::deserialize(data, chunkId, chunkType, status);
+		PacketVersion v;
+		deserializePacketVersionNoHeader(data, v);
+		if (v == cstoma::deleteChunk::kECChunks) {
+			cstoma::deleteChunk::deserialize(data, chunkId, chunkType, status);
+		} else {
+			legacy::ChunkPartType legacy_type;
+			cstoma::deleteChunk::deserialize(data, chunkId, legacy_type, status);
+			chunkType = legacy_type;
+		}
 	}
 
 	chunk_got_delete_status(eptr, chunkId, chunkType, status);
@@ -756,15 +796,31 @@ int matocsserv_send_liz_replicatechunk(matocsserventry *eptr, uint64_t chunkid,
 			return 0;
 		}
 	}
-	std::vector<ChunkTypeWithAddress> sources;
-	for (size_t i = 0; i < sourcePointers.size(); ++i) {
-		matocsserventry *src = sourcePointers[i];
-		sources.push_back(ChunkTypeWithAddress(
-				NetworkAddress(src->servip, src->servport), sourceTypes[i]));
+	if (eptr->version < kFirstECVersion) {
+		std::vector<legacy::ChunkTypeWithAddress> sources;
+		for (size_t i = 0; i < sourcePointers.size(); ++i) {
+			sassert((int)sourceTypes[i].getSliceType() < Goal::Slice::Type::kECFirst);
+
+			matocsserventry *src = sourcePointers[i];
+			sources.push_back(legacy::ChunkTypeWithAddress(
+			    NetworkAddress(src->servip, src->servport), (legacy::ChunkPartType)sourceTypes[i]));
+		}
+		eptr->outputPackets.push_back(OutputPacket());
+		matocs::replicateChunk::serialize(eptr->outputPackets.back().packet, chunkid, version,
+		                                  (legacy::ChunkPartType)type, sources);
+	} else {
+		std::vector<ChunkTypeWithAddress> sources;
+		for (size_t i = 0; i < sourcePointers.size(); ++i) {
+			matocsserventry *src = sourcePointers[i];
+			sources.push_back(ChunkTypeWithAddress(
+				NetworkAddress(src->servip, src->servport),
+				sourceTypes[i],
+				src->version));
+		}
+		eptr->outputPackets.push_back(OutputPacket());
+		matocs::replicateChunk::serialize(eptr->outputPackets.back().packet, chunkid, version, type,
+		                                  sources);
 	}
-	eptr->outputPackets.push_back(OutputPacket());
-	matocs::replicateChunk::serialize(eptr->outputPackets.back().packet,
-			chunkid, version, type, sources);
 	matocsserv_replication_begin(chunkid, version, type,
 			eptr, sourcePointers.size(), sourcePointers.data());
 	return 0;
@@ -778,7 +834,15 @@ void matocsserv_got_replicatechunk_status(matocsserventry *eptr, const std::vect
 	uint8_t status;
 
 	if (packetType == LIZ_CSTOMA_REPLICATE_CHUNK) {
-		cstoma::replicateChunk::deserialize(data, chunkId, chunkType, status, chunkVersion);
+		PacketVersion v;
+		deserializePacketVersionNoHeader(data, v);
+		if (v == cstoma::replicateChunk::kECChunks) {
+			cstoma::replicateChunk::deserialize(data, chunkId, chunkType, status, chunkVersion);
+		} else {
+			legacy::ChunkPartType legacy_type;
+			cstoma::replicateChunk::deserialize(data, chunkId, legacy_type, status, chunkVersion);
+			chunkType = legacy_type;
+		}
 	} else {
 		sassert(packetType == CSTOMA_REPLICATE);
 		deserializeAllMooseFsPacketDataNoHeader(data, chunkId, chunkVersion, status);
@@ -786,7 +850,7 @@ void matocsserv_got_replicatechunk_status(matocsserventry *eptr, const std::vect
 
 	matocsserv_replication_end(chunkId, chunkVersion, chunkType, eptr);
 	chunk_got_replicate_status(eptr, chunkId, chunkVersion, chunkType, status);
-	if (status != 0) {
+	if (status != 0 && status != LIZARDFS_ERROR_WAITING) {
 		syslog(LOG_NOTICE, "(%s:%" PRIu16 ") chunk: %016" PRIX64 " replication status: %s",
 				eptr->servstrip, eptr->servport, chunkId, mfsstrerr(status));
 	}
@@ -801,8 +865,11 @@ int matocsserv_send_setchunkversion(matocsserventry *eptr, uint64_t chunkId, uin
 			sassert(chunkType == slice_traits::standard::ChunkPartType());
 			serializeMooseFsPacket(eptr->outputPackets.back().packet, MATOCS_SET_VERSION,
 					chunkId, newVersion, chunkVersion);
-		}
-		else {
+		} else if (eptr->version < kFirstECVersion) {
+			sassert((int)chunkType.getSliceType() < Goal::Slice::Type::kECFirst);
+			matocs::setVersion::serialize(eptr->outputPackets.back().packet, chunkId, (legacy::ChunkPartType)chunkType,
+					chunkVersion, newVersion);
+		} else {
 			matocs::setVersion::serialize(eptr->outputPackets.back().packet, chunkId, chunkType,
 					chunkVersion, newVersion);
 		}
@@ -819,7 +886,15 @@ void matocsserv_got_setchunkversion_status(matocsserventry *eptr,
 	if (eptr->version < kFirstXorVersion) {
 		deserializeAllMooseFsPacketDataNoHeader(data, chunkId, status);
 	} else {
-		cstoma::setVersion::deserialize(data, chunkId, chunkType, status);
+		PacketVersion v;
+		deserializePacketVersionNoHeader(data, v);
+		if (v == cstoma::setVersion::kECChunks) {
+			cstoma::setVersion::deserialize(data, chunkId, chunkType, status);
+		} else {
+			legacy::ChunkPartType legacy_type;
+			cstoma::setVersion::deserialize(data, chunkId, legacy_type, status);
+			chunkType = legacy_type;
+		}
 	}
 
 	chunk_got_setversion_status(eptr, chunkId, chunkType, status);
@@ -836,10 +911,15 @@ int matocsserv_send_duplicatechunk(matocsserventry* eptr, uint64_t newChunkId, u
 	}
 
 	OutputPacket outPacket;
-	if (slice_traits::isStandard(chunkType) && eptr->version < kFirstXorVersion) {
+	if (eptr->version < kFirstXorVersion) {
+		sassert(slice_traits::isStandard(chunkType));
 		// Legacy support
 		serializeMooseFsPacket(outPacket.packet, MATOCS_DUPLICATE, newChunkId, newChunkVersion,
 				chunkId, chunkVersion);
+	} else if (eptr->version < kFirstECVersion) {
+		sassert((int)chunkType.getSliceType() < Goal::Slice::Type::kECFirst);
+		matocs::duplicateChunk::serialize(outPacket.packet, newChunkId, newChunkVersion,
+				(legacy::ChunkPartType)chunkType, chunkId, chunkVersion);
 	} else {
 		matocs::duplicateChunk::serialize(outPacket.packet, newChunkId, newChunkVersion,
 				chunkType, chunkId, chunkVersion);
@@ -855,7 +935,15 @@ void matocsserv_got_duplicatechunk_status(matocsserventry* eptr, const std::vect
 	if (eptr->version < kFirstXorVersion) {
 		deserializeAllMooseFsPacketDataNoHeader(data, chunkId, status);
 	} else {
-		cstoma::duplicateChunk::deserialize(data, chunkId, chunkType, status);
+		PacketVersion v;
+		deserializePacketVersionNoHeader(data, v);
+		if (v == cstoma::duplicateChunk::kECChunks) {
+			cstoma::duplicateChunk::deserialize(data, chunkId, chunkType, status);
+		} else {
+			legacy::ChunkPartType legacy_type;
+			cstoma::duplicateChunk::deserialize(data, chunkId, legacy_type, status);
+			chunkType = legacy_type;
+		}
 	}
 
 	chunk_got_duplicate_status(eptr, chunkId, chunkType, status);
@@ -873,13 +961,19 @@ void matocsserv_send_truncatechunk(matocsserventry* eptr, uint64_t chunkid, Chun
 	if (eptr->mode == KILL) {
 		return;
 	}
-	if (slice_traits::isStandard(chunkType) && eptr->version < kFirstXorVersion) {
+	if (eptr->version < kFirstXorVersion) {
+		sassert(slice_traits::isStandard(chunkType));
 		// For MooseFS 1.6.27
 		data = matocsserv_createpacket(eptr,MATOCS_TRUNCATE,8+4+4+4);
 		put64bit(&data,chunkid);
 		put32bit(&data,length);
 		put32bit(&data,newVersion);
 		put32bit(&data,oldVersion);
+	} else if (eptr->version < kFirstECVersion) {
+		sassert((int)chunkType.getSliceType() < (int)kFirstECVersion);
+		eptr->outputPackets.push_back(OutputPacket());
+		matocs::truncateChunk::serialize(eptr->outputPackets.back().packet,
+				chunkid, (legacy::ChunkPartType)chunkType, length, newVersion, oldVersion);
 	} else {
 		eptr->outputPackets.push_back(OutputPacket());
 		matocs::truncateChunk::serialize(eptr->outputPackets.back().packet,
@@ -908,9 +1002,18 @@ void matocsserv_got_truncatechunk_status(matocsserventry *eptr, const uint8_t *d
 void matocsserv_got_liz_truncatechunk_status(matocsserventry *eptr,
 		const std::vector<uint8_t>& data) {
 	uint64_t chunkId;
-	ChunkPartType chunkType = slice_traits::standard::ChunkPartType();
+	ChunkPartType chunkType;
 	uint8_t status;
-	cstoma::truncate::deserialize(data, chunkId, chunkType, status);
+	PacketVersion v;
+
+	deserializePacketVersionNoHeader(data, v);
+	if (v == cstoma::truncate::kECChunks) {
+		cstoma::truncate::deserialize(data, chunkId, chunkType, status);
+	} else {
+		legacy::ChunkPartType legacy_type;
+		cstoma::truncate::deserialize(data, chunkId, legacy_type, status);
+		chunkType = legacy_type;
+	}
 
 	chunk_got_truncate_status(eptr, chunkId, chunkType, status);
 	if (status!=0) {
@@ -927,10 +1030,15 @@ int matocsserv_send_duptruncchunk(matocsserventry* eptr, uint64_t newChunkId, ui
 	}
 
 	OutputPacket outPacket;
-	if (slice_traits::isStandard(chunkType) && eptr->version < kFirstXorVersion) {
+	if (eptr->version < kFirstXorVersion) {
+		sassert(slice_traits::isStandard(chunkType));
 		// Legacy support
 		serializeMooseFsPacket(outPacket.packet, MATOCS_DUPTRUNC, newChunkId, newChunkVersion,
 				chunkId, chunkVersion, newChunkLength);
+	} else if (eptr->version < kFirstECVersion) {
+		sassert((int)chunkType.getSliceType() < Goal::Slice::Type::kECFirst);
+		matocs::duptruncChunk::serialize(outPacket.packet, newChunkId,
+				newChunkVersion, (legacy::ChunkPartType)chunkType, chunkId, chunkVersion, newChunkLength);
 	} else {
 		matocs::duptruncChunk::serialize(outPacket.packet, newChunkId,
 				newChunkVersion, chunkType, chunkId, chunkVersion, newChunkLength);
@@ -946,7 +1054,15 @@ void matocsserv_got_duptruncchunk_status(matocsserventry* eptr, const std::vecto
 	if (eptr->version < kFirstXorVersion) {
 		deserializeAllMooseFsPacketDataNoHeader(data, chunkId, status);
 	} else {
-		cstoma::duptruncChunk::deserialize(data, chunkId, chunkType, status);
+		PacketVersion v;
+		deserializePacketVersionNoHeader(data, v);
+		if (v == cstoma::duptruncChunk::kECChunks) {
+			cstoma::duptruncChunk::deserialize(data, chunkId, chunkType, status);
+		} else {
+			legacy::ChunkPartType legacy_type;
+			cstoma::duptruncChunk::deserialize(data, chunkId, legacy_type, status);
+			chunkType = legacy_type;
+		}
 	}
 
 	chunk_got_duptrunc_status(eptr, chunkId, chunkType, status);
@@ -1210,12 +1326,18 @@ void matocsserv_liz_register_chunks(matocsserventry *eptr, const std::vector<uin
 		throw (IncorrectDeserializationException) {
 	PacketVersion v;
 	deserializePacketVersionNoHeader(data, v);
-	if (v == cstoma::registerChunks::kStandardAndXorChunks) {
+	if (v == cstoma::registerChunks::kECChunks) {
 		std::vector<ChunkWithVersionAndType> chunks;
 		cstoma::registerChunks::deserialize(data, chunks);
 		for (auto& chunk : chunks) {
 			chunk_server_has_chunk(eptr, chunk.id, chunk.version, chunk.type);
 		}
+	} else if (v == cstoma::registerChunks::kStandardAndXorChunks) {
+		std::vector<legacy::ChunkWithVersionAndType> chunks;
+		cstoma::registerChunks::deserialize(data, chunks);
+		for (auto& chunk : chunks) {
+			chunk_server_has_chunk(eptr, chunk.id, chunk.version, ChunkPartType(chunk.type));
+		}
 	} else {
 		std::vector<ChunkWithVersion> chunks;
 		cstoma::registerChunks::deserialize(data, chunks);
@@ -1276,10 +1398,20 @@ void matocsserv_chunk_damaged(matocsserventry *eptr,const uint8_t *data,uint32_t
 }
 
 void matocsserv_liz_chunk_damaged(matocsserventry *eptr, const std::vector<uint8_t>& data) {
-	std::vector<ChunkWithType> chunks;
-	cstoma::chunkDamaged::deserialize(data, chunks);
-	for (const auto& chunk : chunks) {
-		chunk_damaged(eptr, chunk.id, chunk.type);
+	PacketVersion v;
+	deserializePacketVersionNoHeader(data, v);
+	if (v == cstoma::chunkDamaged::kECChunks) {
+		std::vector<ChunkWithType> chunks;
+		cstoma::chunkDamaged::deserialize(data, chunks);
+		for (const auto& chunk : chunks) {
+			chunk_damaged(eptr, chunk.id, chunk.type);
+		}
+	} else {
+		std::vector<legacy::ChunkWithType> chunks;
+		cstoma::chunkDamaged::deserialize(data, chunks);
+		for (const auto& chunk : chunks) {
+			chunk_damaged(eptr, chunk.id, ChunkPartType(chunk.type));
+		}
 	}
 }
 
@@ -1308,10 +1440,20 @@ void matocsserv_chunks_lost(matocsserventry *eptr,const uint8_t *data,uint32_t l
 }
 
 void matocsserv_liz_chunks_lost(matocsserventry *eptr, const std::vector<uint8_t>& data) {
-	std::vector<ChunkWithType> chunks;
-	cstoma::chunkLost::deserialize(data, chunks);
-	for (const auto& chunk : chunks) {
-		chunk_lost(eptr, chunk.id, chunk.type);
+	PacketVersion v;
+	deserializePacketVersionNoHeader(data, v);
+	if (v == cstoma::chunkLost::kECChunks) {
+		std::vector<ChunkWithType> chunks;
+		cstoma::chunkLost::deserialize(data, chunks);
+		for (const auto& chunk : chunks) {
+			chunk_lost(eptr, chunk.id, chunk.type);
+		}
+	} else {
+		std::vector<legacy::ChunkWithType> chunks;
+		cstoma::chunkLost::deserialize(data, chunks);
+		for (const auto& chunk : chunks) {
+			chunk_lost(eptr, chunk.id, ChunkPartType(chunk.type));
+		}
 	}
 }
 
@@ -1337,10 +1479,20 @@ void matocsserv_chunks_new(matocsserventry *eptr,const uint8_t *data,uint32_t le
 }
 
 void matocsserv_liz_chunk_new(matocsserventry *eptr, const std::vector<uint8_t>& data) {
-	std::vector<ChunkWithVersionAndType> chunks;
-	cstoma::chunkNew::deserialize(data, chunks);
-	for (auto& chunk : chunks) {
-		chunk_server_has_chunk(eptr, chunk.id, chunk.version, chunk.type);
+	PacketVersion v;
+	deserializePacketVersionNoHeader(data, v);
+	if (v == cstoma::chunkNew::kECChunks) {
+		std::vector<ChunkWithVersionAndType> chunks;
+		cstoma::chunkNew::deserialize(data, chunks);
+		for (auto& chunk : chunks) {
+			chunk_server_has_chunk(eptr, chunk.id, chunk.version, chunk.type);
+		}
+	} else {
+		std::vector<legacy::ChunkWithVersionAndType> chunks;
+		cstoma::chunkNew::deserialize(data, chunks);
+		for (auto& chunk : chunks) {
+			chunk_server_has_chunk(eptr, chunk.id, chunk.version, ChunkPartType(chunk.type));
+		}
 	}
 }
 
@@ -1471,6 +1623,9 @@ void matocsserv_term(void) {
 }
 
 void matocsserv_read(matocsserventry *eptr) {
+	SignalLoopWatchdog watchdog;
+
+	watchdog.start();
 	for (;;) {
 		uint32_t bytesToRead = eptr->inputPacket.bytesToBeRead();
 		ssize_t ret = read(eptr->sock, eptr->inputPacket.pointerToBeReadInto(), bytesToRead);
@@ -1503,14 +1658,20 @@ void matocsserv_read(matocsserventry *eptr) {
 
 		matocsserv_gotpacket(eptr, eptr->inputPacket.getHeader(), eptr->inputPacket.getData());
 		eptr->inputPacket.reset();
+
+		if (watchdog.expired()) {
+			break;
+		}
 	}
 }
 
 void matocsserv_write(matocsserventry *eptr) {
-	int32_t i;
+	SignalLoopWatchdog watchdog;
+
+	watchdog.start();
 	while (!eptr->outputPackets.empty()) {
 		OutputPacket& pack = eptr->outputPackets.front();
-		i = write(eptr->sock, pack.packet.data() + pack.bytesSent,
+		ssize_t i = write(eptr->sock, pack.packet.data() + pack.bytesSent,
 				pack.packet.size() - pack.bytesSent);
 		if (i<0) {
 			if (errno!=EAGAIN) {
@@ -1524,6 +1685,10 @@ void matocsserv_write(matocsserventry *eptr) {
 			return;
 		}
 		eptr->outputPackets.pop_front();
+
+		if (watchdog.expired()) {
+			break;
+		}
 	}
 }
 
